from typing import Generic, TypeVar, Optional, Union, List, Dict, Any
from pydantic.generics import GenericModel, BaseModel

_T = TypeVar('_T')


class BaseApiOut(GenericModel, Generic[_T], BaseModel):
    message: str = 'success'
    data: Optional[_T] = None
    code: int = 200


class ItemListSchema(GenericModel, Generic[_T], BaseModel):
    """数据查询返回格式"""
    items: Optional[List[_T]]  # 数据列表
    total: int = None  # 数据总量
    query: Dict[str, Any] = None
    filter: Dict[str, Any] = None


class ListApiOut(BaseApiOut, Generic[_T]):
    data: Union[Optional[ItemListSchema[_T]], List]
    # data: ItemListSchema[_T]


class ErrorApiOut(BaseApiOut, Generic[_T]):
    msg: str = 'error'
    data: Optional[_T] = None
    code: int = 201


class HttpStatus:
    # 请求正常
    HTTP_200_OK = 200

    # 用户登录异常
    HTTP_418_AUTH_EXCEPT = 418

    # 用户不存在
    HTTP_419_USER_EXCEPT = 419

    # Token 过期
    HTTP_420_TOKEN_EXCEPT = 420

    # 内部参数校验失败
    HTTP_421_INNER_PARAM_EXCEPT = 421

    # 角色不存在
    HTTP_422_ROLE_NOT_EXIST = 422

    # 请求参数格式错误
    HTTP_422_QUERY_PARAM_EXCEPT = 422

    # Authentication 权限异常
    HTTP_425_AUTHENTICATION_EXCEPT = 425

    # 服务端错误
    HTTP_500_INTERNAL_SERVER_ERROR = 500

    # 角色不存在
    HTTP_600_ROLE_NOT_EXIST = 600

    # 角色不存在
    HTTP_601_ROLE_EXIST = 601


__all__ = [
    'BaseApiOut',
    'ItemListSchema',
    'ListApiOut',
    'ErrorApiOut',
    'HttpStatus'
]



# -*- coding: utf-8 -*-
""" 
@author: moxiaoying
@create: 2022/10/16
@description: 基础模型
"""
from tortoise import fields

from typing import Type
from tortoise.models import Model
from tortoise.contrib.pydantic import pydantic_model_creator


class BaseCrudMixin(Model):
    @classmethod
    async def create_one(cls, item: dict):
        return await cls.create(**item)

    @classmethod
    async def find_by(cls, **kwargs):
        return await cls.filter(**kwargs).all()

    @classmethod
    async def find_one(cls, **kwargs):
        return await cls.filter(**kwargs).first()

    @classmethod
    async def update_one(cls, _id: str, item: dict):
        await cls.filter(id=_id).update(**item)
        return cls.get(id=_id)

    @classmethod
    async def delete_one(cls, _id: str) -> int:
        deleted_count = await cls.filter(id=_id).delete()
        return deleted_count

    @classmethod
    async def delete_many(cls, ids: list) -> int:
        deleted_count = await cls.filter(id__in=ids).delete()
        return deleted_count


class BaseSchemaMixin:
    @classmethod
    def base_schema(cls: Type[Model], name, include=(), exclude=(), **kwargs):
        name = f'{cls.__name__}Schema{name}'
        # include = kwargs.get('include', ())
        if include:
            return pydantic_model_creator(cls, name=name, include=include, **kwargs)
        return pydantic_model_creator(cls, name=name, exclude=exclude, **kwargs)

    @classmethod
    def schema_list(cls, name='List', include=(), exclude=(), **kwargs):
        return cls.base_schema(name=name, include=include, exclude=exclude, **kwargs)

    @classmethod
    def schema_create(cls, name='Create', include=(), exclude=(), **kwargs):
        return cls.base_schema(name, include=include, exclude=exclude, exclude_readonly=True, **kwargs)

    @classmethod
    def schema_read(cls, name='Read', include=(), exclude=(), **kwargs):
        return cls.base_schema(name, include=include, exclude=exclude, **kwargs)

    @classmethod
    def schema_update(cls, name='Update', include=(), exclude=(), **kwargs):
        return cls.base_schema(name, include=include, exclude=exclude, exclude_readonly=True, **kwargs)

    @classmethod
    def schema_filters(cls, name='Filters', include=('',), exclude=(), **kwargs):
        return cls.base_schema(name, include=include, exclude=exclude, optional=include, exclude_readonly=True,
                               **kwargs)

    @classmethod
    def schema_delete(cls):
        return int


class TimestampMixin:
    created = fields.DatetimeField(
        null=True, auto_now_add=True, description="创建时间")
    updated = fields.DatetimeField(
        null=True, auto_now=True, description="更新时间")


class AbstractBaseModel(Model):
    id = fields.IntField(pk=True)
    is_delete = fields.IntField(
        null=False, default=0, description="逻辑删除:0=未删除,1=删除")

    class Meta:
        abstract = True


class BaseModel(BaseCrudMixin, BaseSchemaMixin, TimestampMixin, AbstractBaseModel):
    pass

    class PydanticMeta:
        backward_relations = False
        exclude = ('is_delete',)

    class Meta:
        abstract = True
        ordering = ['-updated', '-created']


__all__ = [
    'BaseModel'
]



from enum import Enum
from typing import Union, Any, List
from fastapi import APIRouter, Body, Request, Depends
from tortoise.queryset import QuerySet

from .model import BaseModel
from .response_code import ListApiOut, BaseApiOut, ItemListSchema


class Order(Enum):
    asc: str = ''
    desc: str = '-'


class ModelCrud(APIRouter):
    def __init__(self, model: Union[BaseModel, Any],
                 schema_list=None,
                 schema_create=None,
                 schema_read=None,
                 schema_update=None,
                 schema_delete=None,
                 schema_filters=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        self.schema_list = schema_list or model.schema_list()
        self.schema_read = schema_read or model.schema_read()
        self.schema_update = schema_update or model.schema_update()
        self.schema_create = schema_create or model.schema_create()
        self.schema_delete = schema_delete or model.schema_delete()
        self.schema_filters = schema_filters or model.schema_filters()
        self.register_crud()

    @classmethod
    def pre_create(cls, item: dict) -> dict:
        return item

    @classmethod
    def pre_create_all(cls, items: list):
        for item in items:
            yield cls.pre_create(item.dict())

    @classmethod
    def pre_update(cls, item: dict) -> dict:
        return item

    @classmethod
    def pre_list(cls, queryset: QuerySet, item: dict) -> QuerySet:
        """
        数据预处理：搜索字段
        :param queryset:
        :param item:
        :return:
        """
        filter_field = {f'{k}__icontains': v for k, v in item.items() if v}
        return queryset.filter(**filter_field)

    def register_crud(self):
        # model_name =
        schema_create = self.schema_create
        schema_update = self.schema_update
        schema_filters = self.schema_filters

        @self.post('/list', response_model=ListApiOut[self.schema_list])
        async def handle_list(request: Request, filters: schema_filters, page: int = 1, pagesize: int = 10, sort='id',
                              order: Order = Order.desc, ):
            data = ItemListSchema(items=[])

            queryset = self.model.filter(is_delete=False)
            queryset = self.pre_list(queryset, filters.dict())
            queryset = queryset.order_by(f'{order.value}{sort}')
            data.items = await self.schema_list.from_queryset(queryset.offset((page - 1) * pagesize).limit(pagesize))
            data.total = await queryset.count()
            data.query = request.query_params
            data.filter = filters
            return ListApiOut(data=data)

        @self.get('/read/{item_id}', response_model=BaseApiOut[self.schema_read])
        async def handle_read(item_id):
            data = await self.model.find_one(id=item_id)
            data = await self.schema_read.from_tortoise_orm(data)
            return BaseApiOut(data=data)

        @self.post("/create", response_model=BaseApiOut[self.schema_create])
        async def handle_create(item: schema_create):
            item = self.pre_create(item.dict())
            new_item = await self.model.create_one(item)
            data = await self.schema_create.from_tortoise_orm(new_item)
            return BaseApiOut(data=data)

        @self.post('/create/all', response_model=BaseApiOut)
        async def handle_create_all(items: List[schema_create]):
            items = self.pre_create_all(items)
            await self.model.bulk_create([self.model(**item) for item in items], ignore_conflicts=True)
            return BaseApiOut(msg='批量创建成功')
            pass

        @self.put("/{item_id}", response_model=BaseApiOut[schema_update])
        async def handle_update(item_id: str, item: schema_update):
            item = self.pre_update(item.dict(exclude_unset=True))
            updated_item = await self.model.update_one(item_id, item)
            data = await schema_update.from_queryset_single(updated_item)
            return BaseApiOut(data=data)

        @self.delete("/{item_ids}", description='删除1条或多条数据example：1,2',
                     response_model=BaseApiOut[self.schema_delete])
        async def handle_delete(item_ids: str):
            ids = item_ids.split(',')
            data = await self.model.delete_many(ids)
            return BaseApiOut(data=data)

        # @self.post('/search', response_model=ListApiOut[schema_filters])
        # async def handle_search(filters: schema_filters):
        #     data = ItemListSchema(items=[])
        #     queryset = self.model.filter(is_delete=False, **filters.dict())
        #     data.items = await schema_filters.from_queryset(queryset.all())
        #     data.total = await queryset.count()
        #     data.filter = filters
        #     return ListApiOut(data=data)

        return self


__all__ = [
    'ModelCrud'
]
